/* -*-c-*-
 *
 * Protocol specific definitions for Unix-domain sockets
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of the `fwd' port forwarder.
 *
 * `fwd' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * `fwd' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with `fwd'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "fwd.h"

/*----- Data structures ---------------------------------------------------*/

typedef struct un_addr {
  addr a;
  struct sockaddr_un sun;
} un_addr;

typedef struct un_opts {
  addr_opts ao;
  fattr f;
} un_opts;

/*----- Protocol operations -----------------------------------------------*/

/* --- @read@ --- */

static addr *un_read(scanner *sc, unsigned type)
{
  dstr d = DSTR_INIT;
  un_addr *ua;

  conf_fname(sc, &d);
  ua = xmalloc(sizeof(addr) +
	       offsetof(struct sockaddr_un, sun_path) +
	       d.len + 1);
  ua->a.ops = &un_ops;
  ua->a.sz = offsetof(struct sockaddr_un, sun_path) + d.len + 1;
  memset(&ua->sun, 0, ua->a.sz);
  ua->sun.sun_family = AF_UNIX;
  memcpy(ua->sun.sun_path, d.buf, d.len + 1);
  dstr_destroy(&d);
  return (&ua->a);
}

/* --- @destroy@ --- */

static void un_destroy(addr *a)
{
  un_addr *ua = (un_addr *)a;
  xfree(ua);
}

/* --- @print@ --- */

static void un_print(addr *a, unsigned type, dstr *d)
{
  un_addr *ua = (un_addr *)a;
  dstr_puts(d, "unix:");
  dstr_puts(d, ua->sun.sun_path);
}

/* --- @initopts@ --- */

static addr_opts *un_initopts(void)
{
  un_opts *uo = CREATE(un_opts);
  uo->f = fattr_global;
  return (&uo->ao);
}

/* --- @option@ --- */

static int srcopt(scanner *sc, addr_opts *ao)
{
  un_opts *uo = (un_opts *)ao;
  CONF_BEGIN(sc, "source", "Unix domain socket source")

  if (fattr_option(sc, uo ? &uo->f : &fattr_global))
    CONF_ACCEPT;

  CONF_END;
}

static int un_option(scanner *sc, addr_opts *ao, unsigned type)
{
  CONF_BEGIN(sc, "unix", "Unix domain socket");
  if (type != ADDR_DEST && srcopt(sc, ao))
    CONF_ACCEPT;
  CONF_END;
}

/* --- @accept@ --- */

static reffd *un_accept(int fd, addr_opts *ao, const char *desc)
{
  int nfd;
  un_opts *uo = (un_opts *)ao;

  /* --- Accept the new connection --- */

  {
    char buf[PATH_MAX + sizeof(struct sockaddr)];
    struct sockaddr_un *sun = (struct sockaddr_un *)buf;
    socklen_t sunsz = sizeof(buf);

    if ((nfd = accept(fd, (struct sockaddr *)sun, &sunsz)) < 0)
      return (0);
  }

  /* --- Log the connection --- *
   *
   * It'd be really nice if I could find out who the user is, but I can't in
   * anything like a portable way.
   */

  if (!(uo->ao.f & ADDRF_NOLOG))
    fw_log(-1, "[%s] accepted", desc);
  return (reffd_init(nfd));
}

/* --- @freeopts@ --- */

static void un_freeopts(addr_opts *ao)
{
  un_opts *uo = (un_opts *)ao;
  DESTROY(uo);
}

/* --- @bind@ --- */

static int un_bind(addr *a, addr_opts *ao)
{
  un_addr *ua = (un_addr *)a;
  un_opts *uo = (un_opts *)ao;
  int fd;

  if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    goto fail_0;
  if (bind(fd, (struct sockaddr *)&ua->sun, sizeof(ua->sun)))
    goto fail_1;
  if (fattr_apply(ua->sun.sun_path, &uo->f))
    goto fail_1;
  return (fd);

fail_1:
  close(fd);
fail_0:
  return (-1);
}

/* --- @unbind@ --- */

static void un_unbind(addr *a)
{
  un_addr *ua = (un_addr *)a;
  unlink(ua->sun.sun_path);
}

/* --- @connect@ --- */

static int un_connect(addr *a, addr_opts *ao, conn *c, endpt *e)
{
  un_addr *ua = (un_addr *)a;
  int fd;

  if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    goto fail_0;
  return (conn_init(c, sel, fd, (struct sockaddr *)&ua->sun, sizeof(ua->sun),
		    starget_connected, e));
fail_0:
  return (-1);
}

/* --- Protocol definition --- */

addr_ops un_ops = {
  "unix",
  un_read, un_destroy, un_print,
  un_initopts, un_option, 0, un_freeopts, un_bind, un_unbind, un_accept,
  0, 0, un_connect
};

/*----- That's all, folks -------------------------------------------------*/
